import os
import sys
import logging
import subprocess
from typing import List


def run(cmd: List[str]) -> str:
    """Execute external program.

    :params cmd: Command to execute.

    :return: Output from stdout.
    """

    log = logging.getLogger('rich')

    try:
        proc = subprocess.Popen(cmd,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                encoding='utf-8')

        stdout, stderr = proc.communicate()

        if proc.returncode != 0:
            log.error(stderr)
            log.error(f'Return code: {proc.returncode}')
            sys.exit(1)
    except OSError as e:
        print(e)
        log.error(stderr)
        log.error('Failed to execute command.')
        sys.exit(1)

    return stdout


def run_bash(command: str):
    """Execute command via bash."""

    process = subprocess.run(["bash", "-c", command],
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             stdin=None,
                             check=True,
                             encoding='utf-8')

    return process.stdout

def check_dependencies(programs: List[str], exit_on_fail: bool = True):
    """Check if programs are on the system path.

    :param programs: Names of executable programs.
    :param exit_on_fail: Exit program with error code -1 if any program in not on path.

    :return: True if all programs are on path, else False.
    """

    for program in programs:
        if not check_on_path(program, exit_on_fail):
            return False

    return True


def check_on_path(program: str, exit_on_fail: bool = True):
    """Check if program is on the system path.

    :param programs: Names of executable programs.
    :param exit_on_fail: Exit program with error code -1 if any program in not on path.

    :return: True if all program is on path, else False.
    """

    if which(program):
        return True

    if exit_on_fail:
        print(f'{program} is not on the system path.')
        sys.exit(1)

    return False


def which(program: str):
    """Return path to program.

    This is a Python implementation of the linux
    command 'which'.

    http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python

    :param programs: Names of executable programs.

    :return: Path to executable, or None if it isn't on the path.
    """

    fpath, _fname = os.path.split(program)
    if fpath:
        if is_executable(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_executable(exe_file):
                return exe_file

    return None


def is_executable(fpath: str):
    """Check if file is executable.

    :param fpath: Path to file.

    :return: True if executable, else False.
    """

    return os.path.isfile(fpath) and os.access(fpath, os.X_OK)



import multiprocessing as mp
from math import floor
from pathlib import Path
from typing import Collection

import dendropy
import numpy as np
from rich.progress import Progress, TextColumn, BarColumn, TaskProgressColumn, TimeRemainingColumn

from gtdblib import log


def bootstrap_merge_replicates(
        input_tree: Path,
        output_tree: Path,
        replicate_trees: Collection[Path],
        cpus: int = 1
):
    """Calculates non-parametric bootstraps for monophyletic groups and
    transposes them to the input tree.

    :param input_tree: The tree to merge the bootstraps to.
    :param output_tree: The path to write the tree to.
    :param replicate_trees: The trees to calculate the bootstraps from.
    :param cpus: The number of processes to use.
    """

    # Validate arguments
    if not isinstance(input_tree, Path):
        raise ValueError(f'Input tree is not a Path object.')
    if not isinstance(output_tree, Path):
        raise ValueError(f'Output tree is not a Path object.')
    if not isinstance(replicate_trees, Collection):
        raise ValueError(f'Replicate trees is not a Collection object.')
    if not all(isinstance(x, Path) for x in replicate_trees):
        raise ValueError(f'Not all replicate trees are Path objects.')
    if cpus is None or not isinstance(cpus, int) or cpus < 1:
        cpus = 1

    # Load the main tree that support values will be added to
    tree = dendropy.Tree.get_from_path(
        str(input_tree),
        schema='newick',
        rooting='force-unrooted',
        preserve_underscores=True
    )

    # Seed the support values
    for node in tree.internal_nodes():
        node.label = 0
        node.nontrivial_splits = 0

    # Store a consistent ordering of the internal nodes
    input_internal_nodes = tuple(tree.internal_nodes(), )

    # Generate the descendant taxa for those nodes
    input_desc_taxa = tuple([frozenset({y.taxon.label for y in x.leaf_nodes()}) for x in input_internal_nodes])

    # Process the input descendant taxa and
    results = list()
    with Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            TimeRemainingColumn(),
            refresh_per_second=1
    ) as p_bar:
        task_total = p_bar.add_task(
            total=100,
            description=f'Processing tree 1 of {len(replicate_trees):,}'
        )

        # Process each of the trees
        for rep_tree_path in replicate_trees:
            results.append(_calculate_support_worker(
                rep_tree_path,
                input_desc_taxa,
                p_bar,
                task_total,
                len(replicate_trees),
                cpus
            ))
            p_bar.update(
                task_total,
                description=f'Processing tree {len(results):,} of {len(replicate_trees):,}'
            )

    # Append the split information to the nodes
    for result in results:
        for node, (n_support, n_non_trivial_split) in zip(input_internal_nodes, result):
            node.label += n_support
            node.nontrivial_splits += n_non_trivial_split

    # Calculate the %
    for node in tree.internal_nodes():
        if node.nontrivial_splits > 0:
            node.label = str(int(floor(node.label * 100.0 / node.nontrivial_splits)))
        else:
            node.label = 'NA'

    tree.write_to_path(str(output_tree), schema='newick', suppress_rooting=True, unquoted_underscores=True)
    log.info(f'Wrote tree to: {output_tree}')

    return


def _calculate_split_worker(jobs):
    idx, cur_taxa_labels, rep_tree_taxa_set, d_taxon_label_to_bit_idx, all_taxa_bitmask, rep_tree_list = jobs

    # Only calculate the support for splits that are present in the reference tree
    common_taxon_labels = cur_taxa_labels.intersection(rep_tree_taxa_set)

    # Store the number of supported splits
    if len(common_taxon_labels) > 1:

        # Create the bit vector for the taxa
        # This works as the bit shift done (1 << i) will always produce a
        # bit vector with only one significant bit (i.e. the index of the
        # taxon in the taxon namespace)
        split_bit_vec = np.zeros(len(rep_tree_taxa_set), dtype=np.bool)
        split_bit_vec[[d_taxon_label_to_bit_idx[x] for x in common_taxon_labels]] = True

        # Reverse the ordering (as idx=0 should be 001, not 100)
        split_bit_str = ''.join(['1' if x else '0' for x in reversed(split_bit_vec)])

        # Calculate the split support
        split = int(split_bit_str, 2)
        normalized_split = dendropy.Bipartition.normalize_bitmask(
            bitmask=split,
            fill_bitmask=all_taxa_bitmask,
            lowest_relevant_bit=1)

        n_support = int(rep_tree_list.frequency_of_bipartition(split_bitmask=normalized_split))
        n_non_trivial_split = 1
    else:
        n_support = 0
        n_non_trivial_split = 0

    # Save the result
    return idx, n_support, n_non_trivial_split


def _calculate_support_worker(rep_tree_path, taxa_labels, p_bar, task_total, n_trees, cpus):
    """Worker to calculate the support for splits."""

    # Load the tree
    rep_tree = dendropy.Tree.get_from_path(
        str(rep_tree_path),
        schema='newick',
        rooting='force-unrooted',
        preserve_underscores=True
    )
    all_taxa_bitmask = rep_tree.taxon_namespace.all_taxa_bitmask()

    # Calculate the descendant taxa for the reference tree
    rep_tree_list = dendropy.TreeList([rep_tree])
    rep_tree_taxa_set = frozenset({x.taxon.label for x in rep_tree.leaf_node_iter()})

    # Get the bitmask for all taxa that are common
    d_taxon_label_to_bit_idx = {x.label: i for i, x in enumerate(rep_tree.taxon_namespace)}

    # Iterate over the reference tree taxa internal nodes (order is consistent)
    jobs = list()
    for idx, cur_taxa_labels in enumerate(taxa_labels):
        jobs.append((
            idx,
            cur_taxa_labels,
            rep_tree_taxa_set,
            d_taxon_label_to_bit_idx,
            all_taxa_bitmask,
            rep_tree_list
        ))

    # Create a progress bar and process the jobs
    task_inner = p_bar.add_task(f"{rep_tree_path.name}", total=len(jobs))
    d_idx_to_result = dict()
    pct_per_job = 100 / len(jobs) / n_trees

    # Process the jobs single-threaded
    if cpus == 1:
        for result in (_calculate_split_worker(job) for job in jobs):
            d_idx_to_result[result[0]] = result[1:]
            p_bar.update(task_inner, advance=1)
            p_bar.update(task_total, advance=pct_per_job)

    # Process multiple jobs in parallel
    with mp.Pool(processes=cpus) as pool:
        for result in pool.imap_unordered(_calculate_split_worker, jobs, chunksize=10):
            d_idx_to_result[result[0]] = result[1:]
            p_bar.update(task_inner, advance=1)
            p_bar.update(task_total, advance=pct_per_job)

    # Remove the progress bar for this tree
    p_bar.remove_task(task_inner)

    # Return the results (in the same order)
    results = [d_idx_to_result[i] for i in range(len(taxa_labels))]
    return tuple(results, )



import sys
from typing import Dict, List

from gtdblib import log
from gtdblib.taxon.taxon import Taxon
from gtdblib.taxon.rank import TaxonRank
from gtdblib.util.types import is_float
from gtdblib.util.shell.gtdbshutil import check_file_exists
from gtdblib.util.bio.accession import canonical_gid


def read_taxonomy(taxonomy_file: str, use_canonical_gid: bool = False) -> Dict[str, List[str]]:
    """Read Greengenes-style taxonomy file.

    This method is generic in that it can read any Greengenes-style taxonomy string
    and does not strictly require there to be the seven canonical ranks present.

    Expected format is:
        <accession>\t<taxonomy string>

    where the taxonomy string will typically have the formats:
        d__; c__; o__; f__; g__; s__

    :param taxonomy_file: File indicating Greengenes-style taxonomic assignments.
    :param use_canonical_gid: Flag indicating if accessions should be converted to their canonical form.
    :return: Mapping from accessions to a list of taxa.
    """

    check_file_exists(taxonomy_file)

    try:
        d = {}
        for row, line in enumerate(open(taxonomy_file)):
            line_split = line.split('\t')
            unique_id = line_split[0]

            if use_canonical_gid:
                unique_id = canonical_gid(unique_id)

            tax_str = line_split[1].rstrip()
            if tax_str[-1] == ';':
                # remove trailing semicolons which sometimes
                # appear in Greengenes-style taxonomy files
                tax_str = tax_str[0:-1]

            d[unique_id] = [x.strip() for x in tax_str.split(';')]
    except:
        log.error(f'Failed to parse taxonomy file on line {row+1}.')
        raise

    return d


def fill_trailing_ranks(taxa: List[str]) -> List[str]:
    """Fill in missing trailing ranks in a taxonomy string.

    This function assumes it will be provided with a list
    of taxa from the 7 canonical ranks in rank order. Any
    trailing ranks are filled in.

    Example: [d__Bacteria, d__Firmicutes] => [d__Bacteria, d__Firmicutes, c__, o__, f__, g__, s__]

    :param taxa list: List of taxa.

    :return: List of taxa with filled trailing ranks.
    """

    if not taxa:
        return ';'.join(TaxonRank.RANK_PREFIXES)

    try:
        last_rank = Taxonomy.RANK_PREFIXES.index(taxa[-1][0:3])
    except:
        log.error('Taxon is missing rank prefix: %s' % ';'.join(taxa))
        sys.exit()

    for i in range(last_rank+1, len(Taxonomy.RANK_PREFIXES)):
        taxa.append(Taxonomy.RANK_PREFIXES[i])

    return taxa


def read_taxonomy_from_tree(tree: str, warnings: bool = True) -> Dict[str, List[str]]:
    """Obtain the taxonomy for each extant taxa as specified by internal tree labels.

    This method is generic in that it can read any Greengenes-style taxonomy string
    and does not strictly require extran taxa to be classified to theseven canonical ranks.

    :param tree: Filename of Newick tree or Dendropy tree object.
    :param warnings: Flag indicating if issues reading tree should be logged as warnings.

    :return: Mapping from extent taxon labels to a list of taxa.
    """

    if isinstance(tree, str):
        import dendropy
        tree = dendropy.Tree.get_from_path(tree,
                                           schema='newick',
                                           rooting="force-rooted",
                                           preserve_underscores=True)

    taxonomy = {}
    for leaf in tree.leaf_node_iter():
        taxa = []

        node = leaf.parent_node
        while node:
            if node.label:
                taxa_str = node.label
                if ':' in taxa_str:
                    taxa_str = taxa_str.split(':')[1]

                if not is_float(taxa_str):
                    if taxa_str[-1] == ';':
                        taxa_str = taxa_str[:-1]

                    # check for concatenated ranks of the form: p__Crenarchaeota__c__Thermoprotei
                    for prefix in Taxonomy.rank_prefixes:
                        split_str = '__' + prefix
                        if split_str in taxa_str:
                            taxa_str = taxa_str.replace(
                                split_str, ';' + prefix)

                    # appears to be an internal label and not simply a support value
                    taxa = [x.strip() for x in taxa_str.split(';')] + taxa
            node = node.parent_node

        if warnings and len(taxa) > len(TaxonRank.RANK_LABELS):
            log.warning('Invalid taxonomy string read from tree for taxon {}: {}'.format(
                leaf.taxon.label,
                ';'.join(taxa)))

        # check if genus name should be appended to species label as some trees
        # indicate species only by their specific epithet
        if len(taxa) == len(TaxonRank.RANK_LABELS):
            genus = taxa[5][3:]
            species = taxa[6][3:]
            if genus not in species and len(species.split()) == 1:
                taxa[6] = 's__' + genus + ' ' + species

        taxa = fill_trailing_ranks(taxa)
        taxonomy[leaf.taxon.label] = taxa

    return taxonomy


class Taxonomy:
    __slots__ = ('d', 'p', 'c', 'o', 'f', 'g', 's')

    RANK_PREFIXES = ('d__', 'p__', 'c__', 'o__', 'f__', 'g__', 's__')
    RANK_LABELS = ('domain', 'phylum', 'class', 'order',
                   'family', 'genus', 'species')
    RANK_INDEX = {'d__': 0, 'p__': 1, 'c__': 2,
                  'o__': 3, 'f__': 4, 'g__': 5, 's__': 6}

    DOMAIN_INDEX = 0
    PHYLUM_INDEX = 1
    CLASS_INDEX = 2
    ORDER_INDEX = 3
    FAMILY_INDEX = 4
    GENUS_INDEX = 5
    SPECIES_INDEX = 6

    def __init__(self, d: Taxon, p: Taxon, c: Taxon, o: Taxon, f: Taxon, g: Taxon, s: Taxon):
        self.d = d
        self.p = p
        self.c = c
        self.o = o
        self.f = f
        self.g = g
        self.s = s



import re

RE_CANONICAL = re.compile(r'^(?:GB_)?(?:RS_)?(?:GCF_)?(?:GCA_)?(\d{9})\.\d')


def canonical_gid(gid: str) -> str:
    """Get canonical form of NCBI genome accession.

    Example:
        G005435135 -> G005435135
        GCF_005435135.1 -> G005435135
        GCF_005435135.1_ASM543513v1_genomic -> G005435135
        RS_GCF_005435135.1 -> G005435135
        GB_GCA_005435135.1 ->

    :param gid: Genome accesion to conver to canonical form.
    :return: Canonical form of accession.
    """

    match = RE_CANONICAL.match(gid)
    if match:
        return f'G{match[1]}'
    else:
        return gid


def is_same_accn_version(accn1: str, accn2: str) -> bool:
    """Check if accessions have same version number.

    This method assumes accession versions are provided
    as a suffix proceeding a period. This is the format
    used for NCBI accessions, e.g. GCF_005435135.1.

    :param accn1: First accession.
    :param accn2: Second accession.
    :return: True if version number is the same, else False.
    """

    return int(accn1.split('.')[1]) == int(accn2.split('.')[1])



"""Methods for validating a Greengenes-style taxonomy.

This class validates canonical 7 rank Greengenes-style taxonomy string:
    d__; c__; o__; f__; g__; s__
"""

import re
from collections import defaultdict
from typing import Dict, List, Tuple

from gtdblib import log
from gtdblib.taxon.rank import TaxonRank


def validate_taxonomy(
        taxonomy: Dict[str, List[str]],
        check_prefixes: bool,
        check_ranks: bool,
        check_hierarchy: bool,
        check_species: bool,
        check_group_names: bool,
        check_duplicate_names: bool,
        check_capitalization: bool,
        report_errors: bool = True) -> Tuple[Dict[str, str], Dict[str, str], Dict[str, List[str]], Dict[str, List[str]]]:
    """Check if taxonomy forms a strict hierarchy with all expected ranks.

    This method implements a full workflow for validating Greengenes-style taxonomy strings assigned
    to a set of accessions. It identifies a number of common issues and the issues checked can be
    configured using the set of boolean inputs.

    Parameters
    ----------
    taxonomy : d[unique_id] -> [d__<taxon>; ...; s__<taxon>]
        Taxonomy strings indexed by unique ids.
    check_prefixes : boolean
        Flag indicating if prefix of taxon should be validated.
    check_ranks : boolean
        Flag indicating if the presence of all ranks should be validated.
    check_hierarchy : boolean
        Flag indicating if the taxonomic hierarchy should be validated.
    check_species : boolean
        Flag indicating if the taxonomic consistency of named species should be validated.
    check_group_names : boolean
        Flag indicating if group names should be checked for invalid characters.
    check_duplicate_names : boolean
        Flag indicating if group names should be checked for duplicates.
    report_errors : boolean
        Flag indicating if errors should be written to screen.

    Returns
    -------
    dict : d[taxon_id] -> taxonomy
        Taxa with invalid number of ranks.
    dict : d[taxon_id] -> [taxon, taxonomy]
        Taxa with invalid rank prefixes.
    dict: d[taxon_id] -> [species name, error message]
        Taxa with invalid species names.
    dict: d[child_taxon_id] -> two or more parent taxon ids
        Taxa with invalid hierarchies.
    """

    # check for incomplete taxonomy strings or unexpected rank prefixes
    try:
        invalid_ranks = {}
        invalid_prefixes = {}
        invalid_species_name = {}
        invalid_group_name = {}
        invalid_capitalization = set()
        for taxon_id, taxa in taxonomy.items():
            if check_ranks:
                if len(taxa) != len(TaxonRank.RANK_PREFIXES):
                    invalid_ranks[taxon_id] = ';'.join(taxa)
                    continue

            if check_prefixes:
                for r, taxon in enumerate(taxa):
                    if taxon[0:3] != TaxonRank.RANK_PREFIXES[r]:
                        invalid_prefixes[taxon_id] = [taxon, ';'.join(taxa)]
                        break

            if check_group_names:
                for taxon in taxa:
                    canonical_taxon = ' '.join(
                        [t.strip() for t in re.split('_[A-Z]+(?= |$)', taxon[3:])]).strip()
                    if canonical_taxon and re.match('^[a-zA-Z0-9- ]+$', canonical_taxon) is None:
                        if not taxon.startswith('s__') or check_species:
                            invalid_group_name[taxon_id] = [
                                taxon, 'Taxon contains invalid characters']

            if check_species:
                genus_index = TaxonRank.RANK_INDEX['g__']
                species_index = TaxonRank.RANK_INDEX['s__']
                if len(taxa) > species_index:
                    species_name = taxa[species_index]
                    valid, error_msg = validate_species_name(
                        species_name, require_full=True, require_prefix=True)
                    if not valid:
                        invalid_species_name[taxon_id] = [
                            species_name, error_msg]

                    if species_name != 's__':
                        genus_name = taxa[genus_index]
                        generic_name = species_name.split()[0]
                        if genus_name[3:] != generic_name[3:]:
                            invalid_species_name[taxon_id] = [
                                species_name, 'Genus and generic names do not match: %s' % genus_name]

            if check_capitalization:
                for taxon in taxa:
                    if len(taxon) > 3 and taxon[3].islower():
                        invalid_capitalization.add(taxon)
    except:
        log.error(
            'Exception raised while processing {} with the taxa set {}'.format(taxon_id, taxa))
        raise

    # check for duplicate names
    invalid_duplicate_name = []
    if check_duplicate_names:
        invalid_duplicate_name = duplicate_names(taxonomy, check_species)

    # check for inconsistencies in the taxonomic hierarchy
    invalid_hierarchies = defaultdict(set)
    missing_parent = set()
    if check_hierarchy:
        expected_parent = taxonomic_consistency(taxonomy, False)

        for taxon_id, taxa in taxonomy.items():
            for r in range(1, len(taxa)):
                if len(taxa[r]) == 3:
                    continue

                if r == TaxonRank.RANK_INDEX['s__'] and not check_species:
                    continue

                if taxa[r] not in expected_parent:
                    missing_parent.add(taxa[r])
                elif taxa[r - 1] != expected_parent[taxa[r]]:
                    invalid_hierarchies[taxa[r]].add(taxa[r - 1])
                    invalid_hierarchies[taxa[r]].add(expected_parent[taxa[r]])

    if report_errors:
        if len(invalid_ranks):
            print('')
            print('Taxonomy contains too few ranks:')
            for taxon_id, taxa_str in invalid_ranks.items():
                print('%s\t%s' % (taxon_id, taxa_str))

        if len(invalid_prefixes):
            print('')
            print('Taxonomy contains an invalid rank prefix:')
            for taxon_id, info in invalid_prefixes.items():
                print('%s\t%s\t%s' % (taxon_id, info[0], info[1]))

        if len(invalid_group_name):
            print('')
            print('Taxa containing invalid characters:')
            for taxon_id, err_msg in invalid_group_name.items():
                print('%s\t%s\t%s' % (taxon_id, err_msg[0], err_msg[1]))

        if len(invalid_species_name):
            print('')
            print('Taxonomy contains invalid species names:')
            for taxon_id, info in invalid_species_name.items():
                print('%s\t%s\t%s' % (taxon_id, info[0], info[1]))

        if len(invalid_duplicate_name):
            print('')
            print('Taxonomy contains identical taxon names in multiple lineages:')
            for duplicate_name in invalid_duplicate_name.keys():
                print('%s' % duplicate_name)

        if len(missing_parent):
            print('')
            print('Taxonomy contains taxa with an undefined parent:')
            for taxon in missing_parent:
                print('%s' % taxon)

        if len(invalid_hierarchies):
            print('')
            print('Taxonomy contains taxa with multiple parents:')
            for child_taxon, parent_taxa in invalid_hierarchies.items():
                print('%s\t%s' % (child_taxon, ', '.join(parent_taxa)))

        if len(invalid_capitalization):
            print('')
            print('Taxa do not start with a capital letter:')
            for taxon in invalid_capitalization:
                print('{}'.format(taxon))

    return invalid_ranks, invalid_prefixes, invalid_species_name, invalid_hierarchies, invalid_group_name, invalid_capitalization


def duplicate_names(taxonomy: Dict[str, List[str]], check_species: bool = True) -> Dict[str, List[str]]:
    """Identify duplicate names in taxonomy.

    Parameters
    ----------
    taxonomy : d[unique_id] -> [d__<taxon>; ...; s__<taxon>]
        Taxonomy strings indexed by unique ids.

    Returns
    -------
    dict : d[taxon] -> lineages
        List of lineages for duplicate taxa.
    """

    # get lineages for each taxon name
    taxon_lineages = defaultdict(set)
    for taxa in taxonomy.values():
        for i, taxon in enumerate(taxa):
            if len(taxon) > 3:
                taxon_lineages[taxon].add(';'.join(taxa[0:i+1]))

    # identify taxon belonging to multiple lineages
    duplicates = {}
    for taxon, lineages in taxon_lineages.items():
        if len(lineages) >= 2:
            if not taxon.startswith('s__') or check_species:
                duplicates[taxon] = lineages

    return duplicates


def validate_species_name(species_name: str, require_full: bool = True, require_prefix: bool = True) -> Tuple[bool, str]:
    """Validate species name.

    A full species name should be  binomial and include a 'generic name' (genus) and
    a 'specific epithet' (species), i.e. Escherichia coli. This method
    assumes the two names should be separated by a space.

    Parameters
    ----------
    species_name : str
        Species name to validate
    require_full : boolean
        Flag indicating if species name must include 'generic name and 'specific epithet'.
    require_prefix : boolean
        Flag indicating if name must start with the species prefix ('s__').

    Returns
    -------
    boolean
        True if species name is valid, otherwise False.
    str
        Reason for failing validation, otherwise None.
    """

    if species_name == 's__':
        return True, None

    # test for prefix
    if require_prefix:
        if not species_name.startswith('s__'):
            return False, 'name is missing the species prefix'

    # remove prefix before testing other properties
    test_name = species_name
    if test_name.startswith('s__'):
        test_name = test_name[3:]

    # test for full name
    if require_full:
        if 'candidatus' in test_name.lower():
            if len(test_name.split(' ')) <= 2:
                return False, 'name appears to be missing the generic name'
        else:
            if len(test_name.split(' ')) <= 1:
                return False, 'name appears to be missing the generic name'

    # check for tell-tale signs on invalid species names
    if " bacterium" in test_name.lower():
        return False, "name contains the word 'bacterium'"
    if " archaeon" in test_name.lower():
        return False, "name contains the word 'archaeon'"
    if " archeaon" in test_name.lower():
        return False, "name contains the word 'archeaon'"
    if "-like" in test_name.lower():
        return False, "name contains '-like'"
    if " group " in test_name.lower():
        return False, "name contains 'group'"
    if " symbiont" in test_name.lower():
        return False, "name contains 'symbiont'"
    if " endosymbiont" in test_name.lower():
        return False, "name contains 'endosymbiont'"
    if " taxon" in test_name.lower():
        return False, "name contains 'taxon'"
    if " cluster" in test_name.lower():
        return False, "name contains 'cluster'"
    if " of " in test_name.lower():
        return False, "name contains 'of'"
    if test_name[0].islower():
        return False, 'first letter of name is lowercase'
    if 'sp.' in test_name.lower():
        return False, "name contains 'sp.'"

    return True, None


def taxonomic_consistency(taxonomy: Dict[str, List[str]], report_errors: bool = True) -> Dict[str, str]:
    """Determine taxonomically consistent classification for taxa at each rank.

    Parameters
    ----------
    taxonomy : d[unique_id] -> [d__<taxon>; ...; s__<taxon>]
        Taxonomy strings indexed by unique ids.
    report_errors : boolean
        Flag indicating if errors should be written to screen.

    Returns
    -------
    dict : d[taxa] -> expected parent
        Expected parent taxon for taxa at all taxonomic ranks, or
        None if the taxonomy is inconsistent.
    """

    expected_parent = {}
    for genome_id, taxa in taxonomy.items():
        if taxa[0] == 'd__Viruses' or '[P]' in taxa[0]:
            # *** This is a HACK. It would be far better to enforce
            # a taxonomically consistent taxonomy, but
            # the viral taxonomy at IMG is currently not consistent
            continue

        for r in range(1, len(taxa)):
            if len(taxa[r]) == 3:
                break

            if taxa[r] in expected_parent:
                if report_errors:
                    if taxa[r - 1] != expected_parent[taxa[r]]:
                        log.warning(
                            'Provided taxonomy is not taxonomically consistent.')
                        log.warning('Genome %s indicates the parent of %s is %s.' % (
                            genome_id, taxa[r], taxa[r - 1]))
                        log.warning('The parent of this taxa was previously indicated as %s.' % (
                            expected_parent[taxa[r]]))

            expected_parent[taxa[r]] = taxa[r - 1]

    return expected_parent



from pathlib import Path

import dendropy
from rich.progress import track

from gtdblib import log
from gtdblib.util.bio.accession import canonical_gid


def convert_tree_accessions_to_canonical(input_path: Path, output_path: Path):
    """Convert all accessions a tree to their canonical form."""

    # Validate arguments
    if not isinstance(input_path, Path):
        raise ValueError(f'Input path is not a Path object.')
    if not isinstance(output_path, Path):
        raise ValueError(f'Output path is not a Path object.')

    # Load the main tree that support values will be added to
    log.info(f'Loading input tree: {input_path}')
    tree = dendropy.Tree.get_from_path(
        str(input_path),
        schema='newick',
        preserve_underscores=True
    )

    # Update the labels for each taxon
    labels_seen = dict()
    for taxon in track(tree.taxon_namespace, description='Processing...'):
        gid = canonical_gid(taxon.label)
        if gid in labels_seen:
            log.warning(f'Leaf node {taxon.label} -> {gid} is a duplicate of {labels_seen[gid]}')
        taxon.label = gid
        labels_seen[gid] = taxon.label

    # Save the tree
    tree.write_to_path(str(output_path), schema='newick', suppress_rooting=True, unquoted_underscores=True)
    log.info(f'Wrote tree to: {output_path}')



import unittest
from pathlib import Path

from gtdblib.tree.convert_accession import convert_tree_accessions_to_canonical


class TestTreeBootstrapMerge(unittest.TestCase):

    def test_bootstrap_merge_replicates_local(self):
        path_out = Path('/tmp/bac120_r2072.tree')
        path_ref = Path('/tmp/bac120_r207.tree')

        convert_tree_accessions_to_canonical(path_ref, path_out)



import unittest

from gtdblib.util.bio.accession import canonical_gid


class TestAccession(unittest.TestCase):

    def test_canonical_gid(self):
        self.assertEqual(canonical_gid('NC_000913.3'), 'NC_000913.3')
        self.assertEqual(canonical_gid('G005435135'), 'G005435135')

        self.assertEqual(canonical_gid('GCA_005435135.1'), 'G005435135')
        self.assertEqual(canonical_gid('GCF_005435135.1'), 'G005435135')

        self.assertEqual(canonical_gid('GB_GCA_005435135.1'), 'G005435135')
        self.assertEqual(canonical_gid('RS_GCF_005435135.1'), 'G005435135')

        self.assertEqual(canonical_gid('GCF_005435135.1_ASM543513v1_genomic'), 'G005435135')
        self.assertEqual(canonical_gid('GCA_005435135.1_ASM543513v1_genomic'), 'G005435135')

        self.assertEqual(canonical_gid('GB_GCA_005435135.1_ASM543513v1_genomic'), 'G005435135')
        self.assertEqual(canonical_gid('RS_GCF_005435135.1_ASM543513v1_genomic'), 'G005435135')
